/*
 * Compute the speed of various tasks
 */

#include "config.h"

#include "bloom.h"

#include <openssl/hmac.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <openssl/rand.h>
#include <string.h>
#include <fcntl.h>
#include <cstring>
#include <assert.h>
#include "aftimer.h"

int setsize  = 1000000;
int bufsize  = setsize*20;               // (we're using sha1)
int hashbytes = 20;				// sha1 is 20 bytes
int COUNT = setsize*2;


u_char bloomtest_sha1[20];
u_int bloomtest_size = sizeof(bloomtest_sha1);

void filetest(const char *fn,int fileio)
{
    u_char *absent_array = (u_char *)malloc(bufsize);
    RAND_pseudo_bytes(absent_array,bufsize);
    int false_positives=0;

    printf("Testing bloom filter speed for absent values with file i/o\n");
    NSRLBloom b2;

    if(b2.open(fn,O_RDONLY)){
	perror("Cannot open Bloom filter:");
	exit(1);
    }
    if(fileio){
	b2.fileio = 1;
    }

    if(!b2.query(bloomtest_sha1)){
	printf("*** BLOOMTEST not in Bloom Filter ***\n");
	exit(1);
    }

    aftimer t;
    printf("Checking for absent values; total to check: %d\n",setsize);
    t.start();
    for(int i=0;i<setsize;i++){
	assert(i*hashbytes < bufsize);
	u_char *hash = absent_array + i*hashbytes;
	if(b2.query(hash)){
	    printf("false positive: %d is present but should be\n",i);
	    false_positives++;
	}
    }
    t.stop();
    printf("Time to query via %s: %8.2f sec  ; %8d Ksamples/sec\n",
	   (fileio ? "file I/O" : "memory-mapped I/O"),
	   t.elapsed_seconds(),(int)(COUNT/t.elapsed_seconds())/1000);
    printf("\n");
}

int main(int argc,char **argv)
{
    /* For testing, put in the SHA1 of "BLOOMTEST" */

    EVP_Digest("Bloom Test",6,bloomtest_sha1,&bloomtest_size,EVP_sha1(),0);

    printf("Bloom Filter tests with a 32-bit bloom filter (512MiBytes; 536MBytes)\n");
    const char *fn = 0;

    if(argc>1){
	fn = argv[1];
	printf("Running to %s\n",fn);
    }
    
    struct stat st;
    if(stat(fn,&st)==0){
	filetest(fn,0);
	filetest(fn,1);
	exit(0);
    }


    // Make a million random numbers for testing
    u_char *hasharray = (u_char *)malloc(bufsize);
    NSRLBloom b;

    RAND_pseudo_bytes(hasharray,bufsize);
    {
	aftimer t;
	printf("Creating bloom filter\n");
	b.create(fn,160,32,4,"A test bloom filter in memory");  // fname, hash_size M, k, comment
	printf("Loading bloom filter with %d random values...\n",setsize);
	t.start();
	for(int i=0;i<setsize;i++){
	    u_char *hash = hasharray + i*hashbytes;
	    b.add(hash);
	}
	t.stop();
	printf("Time to load bloom filter: %8.2f sec  ; %8d Kloads/sec\n",
	       t.elapsed_seconds(),(int)(COUNT/t.elapsed_seconds())/1000);
	printf("\n");
	b.add(bloomtest_sha1);
    }


    {

	if(!b.query(bloomtest_sha1)){
	    printf("*** BLOOMTEST not in Bloom Filter ***\n");
	    exit(1);
	}

	aftimer t;
	printf("Checking for present values; total to check: %d\n",setsize);
	t.start();
	for(int i=0;i<setsize;i++){
	    assert(i*hashbytes < bufsize);
	    u_char *hash = hasharray + i*hashbytes;
	    if(!b.query(hash)){
		printf("error: %d is not present. Blooms should have no false negatives.\n",i);
		exit(1);
	    }
	}
	t.stop();
	printf("Time to query present: %8.2f sec  ; %8d Ksamles/sec\n",
	       t.elapsed_seconds(),(int)(COUNT/t.elapsed_seconds())/1000);
	printf("\n");
    }

    u_char *absent_array = (u_char *)malloc(bufsize);
    RAND_pseudo_bytes(absent_array,bufsize);
    int false_positives=0;

    {
	aftimer t;
	printf("Checking for absent values; total to check: %d\n",setsize);
	t.start();
	for(int i=0;i<setsize;i++){
	    assert(i*hashbytes < bufsize);
	    u_char *hash = absent_array + i*hashbytes;
	    if(b.query(hash)){
		printf("false positive: %d is present but should be\n",i);
		false_positives++;
	    }
	}
	t.stop();
	printf("Time to query absent: %8.2f sec  ; %8d Ksamples/sec\n",
	       t.elapsed_seconds(),(int)(COUNT/t.elapsed_seconds())/1000);
	printf("\n");
    }
    free(absent_array);

    if(false_positives) printf("false positives: %d\n",false_positives);

    if(fn){
	filetest(fn,1);
    }
    return(0);
}

